@@ -1,20 +1,21 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2013-2017 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2013-2017 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import math
 
 from xpra.os_util import monotonic_time
-from xpra.util import MutableInteger, envint, envbool
-from xpra.server.window.region import rectangle, add_rectangle, remove_rectangle, merge_all    #@UnresolvedImport
+from xpra.util import envint, envbool
+from xpra.rectangle import rectangle, add_rectangle, remove_rectangle, merge_all    #@UnresolvedImport
 from xpra.log import Logger
 
 sslog = Logger("regiondetect")
 refreshlog = Logger("regionrefresh")
 
 VIDEO_SUBREGION = envbool("XPRA_VIDEO_SUBREGION", True)
+SUBWINDOW_REGION_BOOST = envint("XPRA_SUBWINDOW_REGION_BOOST", 20)
 
 MAX_TIME = envint("XPRA_VIDEO_DETECT_MAX_TIME", 5)
 MIN_EVENTS = envint("XPRA_VIDEO_DETECT_MIN_EVENTS", 20)
@@ -29,7 +30,7 @@ def scoreinout(ww, wh, region, incount, outcount):
     total = incount+outcount
     assert total>0
     #proportion of damage events that are within this region:
-    inregion = float(incount)/total
+    inregion = incount/total
     #devaluate by taking into account the number of pixels in the area
     #so that a large video region only wins if it really
     #has a larger proportion of the pixels
@@ -41,17 +42,18 @@ def scoreinout(ww, wh, region, incount, outcount):
     width = min(ww, region.width)
     height = min(wh, region.height)
     #proportion of pixels in this region relative to the whole window:
-    inwindow = float(width*height) / (ww*wh)
+    inwindow = (width*height) / (ww*wh)
     ratio = inregion / inwindow
     score = 100.0*inregion
     #if the region has at least 35% of updates, boost it with window ratio
     #(capped at 6, and smoothed with sqrt):
     score += max(0, inregion-0.35) * (math.sqrt(min(6, ratio))-1.0) * RATIO_WEIGHT
-    sslog("scoreinout(%i, %i, %s, %i, %i) inregion=%i%%, inwindow=%i%%, ratio=%.1f, score=%i", ww, wh, region, incount, outcount, 100*inregion, 100*inwindow, ratio, score)
+    sslog("scoreinout(%i, %i, %s, %i, %i) inregion=%i%%, inwindow=%i%%, ratio=%.1f, score=%i",
+          ww, wh, region, incount, outcount, 100*inregion, 100*inwindow, ratio, score)
     return max(0, int(score))
 
 
-class VideoSubregion(object):
+class VideoSubregion:
 
     def __init__(self, timeout_add, source_remove, refresh_cb, auto_refresh_delay, supported=False):
         self.timeout_add = timeout_add
@@ -110,7 +112,7 @@ class VideoSubregion(object):
         if self.detection:
             sslog("video region detection is on - the given region may or may not stick")
         if x==0 and y==0 and w==0 and h==0:
-            self.novideoregion()
+            self.novideoregion("empty")
         else:
             self.rectangle = rectangle(x, y, w, h)
 
@@ -134,7 +136,7 @@ class VideoSubregion(object):
             self.refresh_timer = 0
             self.source_remove(rt)
 
-    def get_info(self):
+    def get_info(self) -> dict:
         r = self.rectangle
         info = {
                 "supported" : self.supported,
@@ -181,7 +183,8 @@ class VideoSubregion(object):
     def remove_refresh_region(self, region):
         remove_rectangle(self.refresh_regions, region)
         remove_rectangle(self.nonvideo_regions, region)
-        refreshlog("remove_refresh_region(%s) updated refresh regions=%s, nonvideo regions=%s", region, self.refresh_regions, self.nonvideo_regions)
+        refreshlog("remove_refresh_region(%s) updated refresh regions=%s, nonvideo regions=%s",
+                   region, self.refresh_regions, self.nonvideo_regions)
 
 
     def add_video_refresh(self, region):
@@ -255,7 +258,7 @@ class VideoSubregion(object):
             self.refresh_timer = self.timeout_add(1000, self.refresh)
 
 
-    def novideoregion(self, msg="", *args):
+    def novideoregion(self, msg, *args):
         sslog("novideoregion: "+msg, *args)
         self.rectangle = None
         self.time = 0
@@ -283,14 +286,38 @@ class VideoSubregion(object):
                 rects = new_rects
         return rects
 
-    def identify_video_subregion(self, ww, wh, damage_events_count, last_damage_events, starting_at=0):
+    def identify_video_subregion(self, ww, wh, damage_events_count, last_damage_events, starting_at=0, children=None):
         if not self.enabled or not self.supported:
             self.novideoregion("disabled")
             return
         if not self.detection:
+            if not self.rectangle:
+                return
+            #just update the fps:
+            from_time = max(starting_at, monotonic_time()-MAX_TIME, self.min_time)
+            self.time = monotonic_time()
+            lde = tuple(x for x in tuple(last_damage_events) if x[0]>=from_time)
+            incount = 0
+            for _,x,y,w,h in lde:
+                r = rectangle(x,y,w,h)
+                inregion = r.intersection_rect(self.rectangle)
+                if inregion:
+                    incount += inregion.width*inregion.height
+            elapsed = monotonic_time()-from_time
+            if elapsed<=0:
+                self.fps = 0
+            else:
+                self.fps = int(incount/(self.rectangle.width*self.rectangle.height) / elapsed)
             return
         sslog("%s.identify_video_subregion(..)", self)
-        sslog("identify_video_subregion(%s, %s, %s, %s)", ww, wh, damage_events_count, last_damage_events)
+        sslog("identify_video_subregion%s",
+              (ww, wh, damage_events_count, last_damage_events, starting_at, children))
+
+        children_rects = ()
+        if children:
+            children_rects = tuple(rectangle(x, y, w, h)
+                                   for _xid, x, y, w, h, _border, _depth in children
+                                   if w>=MIN_W and h>=MIN_H)
 
         if damage_events_count < self.set_at:
             #stats got reset
@@ -299,11 +326,13 @@ class VideoSubregion(object):
         rect = self.rectangle
         if rect and (rect.width>ww or rect.height>wh):
             #region is now bigger than the window!
-            return self.novideoregion("window is now smaller than current region")
+            self.novideoregion("window is now smaller than current region")
+            return
         #arbitrary minimum size for regions we will look at:
         #(we don't want video regions smaller than this - too much effort for little gain)
         if ww<MIN_W or wh<MIN_H:
-            return self.novideoregion("window is too small: %sx%s", MIN_W, MIN_H)
+            self.novideoregion("window is too small: %sx%s", MIN_W, MIN_H)
+            return
 
         def update_markers():
             self.counter = damage_events_count
@@ -318,16 +347,19 @@ class VideoSubregion(object):
             slow_region_timeout = 2 + math.log(2+event_count, 1.5)
             if rect and elapsed>=slow_region_timeout:
                 update_markers()
-                return self.novideoregion("too much time has passed (%is for %i total events)", elapsed, event_count)
-            sslog("identify video: waiting for more damage events (%i) counters: %i / %i", event_count, self.counter, damage_events_count)
+                self.novideoregion("too much time has passed (%is for %i total events)", elapsed, event_count)
+                return
+            sslog("identify video: waiting for more damage events (%i) counters: %i / %i",
+                  event_count, self.counter, damage_events_count)
             return
 
         from_time = max(starting_at, monotonic_time()-MAX_TIME, self.min_time)
         #create a list (copy) to work on:
-        lde = [x for x in tuple(last_damage_events) if x[0]>=from_time]
+        lde = tuple(x for x in tuple(last_damage_events) if x[0]>=from_time)
         dc = len(lde)
         if dc<=MIN_EVENTS:
-            return self.novideoregion("not enough damage events yet (%s)", dc)
+            self.novideoregion("not enough damage events yet (%s)", dc)
+            return
         #structures for counting areas and sizes:
         wc = {}
         hc = {}
@@ -337,7 +369,7 @@ class VideoSubregion(object):
         for _,x,y,w,h in lde:
             rects = self.excluded_rectangles(rectangle(x,y,w,h), ww, wh)
             for r in rects:
-                dec.setdefault(r, MutableInteger()).increase()
+                dec[r] = dec.get(r, 0)+1
                 if w>=MIN_W:
                     wc.setdefault(w, dict()).setdefault(x, set()).add(r)
                 if h>=MIN_H:
@@ -363,7 +395,7 @@ class VideoSubregion(object):
         def damaged_ratio(rect):
             if all_damaged:
                 return 1
-            rects = [rect]
+            rects = (rect, )
             for _,x,y,w,h in lde:
                 r = rectangle(x,y,w,h)
                 new_rects = []
@@ -375,8 +407,9 @@ class VideoSubregion(object):
                 rects = new_rects
             not_damaged_pixels = sum((r.width*r.height) for r in rects)
             rect_pixels = rect.width*rect.height
-            #sslog("damaged_ratio: not damaged pixels(%s)=%i, rect pixels(%s)=%i", rects, not_damaged_pixels, rect, rect_pixels)
-            return max(0, min(1, 1.0-float(not_damaged_pixels)/float(rect_pixels)))
+            #sslog("damaged_ratio: not damaged pixels(%s)=%i, rect pixels(%s)=%i",
+            #     rects, not_damaged_pixels, rect, rect_pixels)
+            return max(0, min(1, 1.0-not_damaged_pixels/rect_pixels))
 
         scores = {None : 0}
         def score_region(info, region, ignore_size=0, d_ratio=0):
@@ -399,8 +432,9 @@ class VideoSubregion(object):
             if d_ratio==0:
                 d_ratio = damaged_ratio(region)
             score = int(score * math.sqrt(d_ratio))
-            sslog("testing %12s video region %34s: %3i%% in, %3i%% out, %3i%% of window, damaged ratio=%.2f, score=%2i",
-                  info, region, 100*incount//total, 100*outcount//total, 100*region.width*region.height/ww/wh, d_ratio, score)
+            children_boost = int(region in children_rects)*SUBWINDOW_REGION_BOOST
+            sslog("testing %12s video region %34s: %3i%% in, %3i%% out, %3i%% of window, damaged ratio=%.2f, children_boost=%i, score=%2i",
+                  info, region, 100*incount//total, 100*outcount//total, 100*region.width*region.height/ww/wh, d_ratio, children_boost, score)
             scores[region] = score
             return score
 
@@ -418,9 +452,9 @@ class VideoSubregion(object):
             self.last_scores = scores
             sslog("score(%s)=%s, damaged=%i%%", self.inout, self.score, self.damaged)
 
-        def setnewregion(rect, msg="", *args):
+        def setnewregion(rect, msg, *args):
             rects = self.excluded_rectangles(rect, ww, wh)
-            if len(rects)==0:
+            if not rects:
                 self.novideoregion("no match after removing excluded regions")
                 return
             if len(rects)==1:
@@ -436,6 +470,7 @@ class VideoSubregion(object):
                     return
             if not self.rectangle or self.rectangle!=rect:
                 sslog("setting new region %s: "+msg, rect, *args)
+                sslog(" is child window: %s", rect in children_rects)
                 self.set_at = damage_events_count
                 self.counter = damage_events_count
             if not self.enabled:
@@ -450,7 +485,8 @@ class VideoSubregion(object):
 
         if len(dec)==1:
             rect, count = tuple(dec.items())[0]
-            return setnewregion(rect, "only region damaged")
+            setnewregion(rect, "only region damaged")
+            return
 
         #see if we can keep the region we already have (if any):
         cur_score = 0
@@ -467,7 +503,7 @@ class VideoSubregion(object):
             #ignore small regions:
             if count>min_count and r.width>=MIN_W and r.height>=MIN_H:
                 damage_count[r] = count
-        c = sum([int(x) for x in damage_count.values()])
+        c = sum(int(x) for x in damage_count.values())
         most_damaged = -1
         most_pct = 0
         if c>0:
@@ -476,7 +512,7 @@ class VideoSubregion(object):
             sslog("identify video: most=%s%% damage count=%s", most_pct, damage_count)
             #is there a region that stands out?
             #try to use the region which is responsible for most of the large damage requests:
-            most_damaged_regions = [r for r,v in damage_count.items() if v==most_damaged]
+            most_damaged_regions = tuple(r for r,v in damage_count.items() if v==most_damaged)
             if len(most_damaged_regions)==1:
                 r = most_damaged_regions[0]
                 score = score_region("most-damaged", r, d_ratio=1.0)
@@ -484,9 +520,13 @@ class VideoSubregion(object):
                 if score>120:
                     setnewregion(r, "%s%% of large damage requests, score=%s", most_pct, score)
                     return
-                elif score>=100:
+                if score>=100:
                     scores[r] = score
 
+        #try children windows:
+        for region in children_rects:
+            scores[region] = score_region("child-window", region, 48*48)
+
         #try harder: try combining regions with the same width or height:
         #(some video players update the video region in bands)
         for w, d in wc.items():
@@ -494,7 +534,7 @@ class VideoSubregion(object):
                 if len(regions)>=2:
                     #merge regions of width w at x
                     min_count = max(2, len(regions)//25)
-                    keep = [r for r in regions if int(dec.get(r, 0))>=min_count]
+                    keep = tuple(r for r in regions if int(dec.get(r, 0))>=min_count)
                     sslog("vertical regions of width %i at %i with at least %i hits: %s", w, x, min_count, keep)
                     if keep:
                         merged = merge_all(keep)
@@ -504,7 +544,7 @@ class VideoSubregion(object):
                 if len(regions)>=2:
                     #merge regions of height h at y
                     min_count = max(2, len(regions)//25)
-                    keep = [r for r in regions if int(dec.get(r, 0))>=min_count]
+                    keep = tuple(r for r in regions if int(dec.get(r, 0))>=min_count)
                     sslog("horizontal regions of height %i at %i with at least %i hits: %s", h, y, min_count, keep)
                     if keep:
                         merged = merge_all(keep)
@@ -514,19 +554,22 @@ class VideoSubregion(object):
         highscore = max(scores.values())
         #a score of 100 is neutral
         if highscore>=120:
-            region = [r for r,s in scores.items() if s==highscore][0]
-            return setnewregion(region, "very high score: %s", highscore)
+            region = next(iter(r for r,s in scores.items() if s==highscore))
+            setnewregion(region, "very high score: %s", highscore)
+            return
 
         #retry existing region, tolerate lower score:
         if cur_score>=90 and (highscore<100 or cur_score>=highscore):
             sslog("keeping existing video region %s with score %s", rect, cur_score)
-            return setnewregion(self.rectangle, "existing region with score: %i" % cur_score)
+            setnewregion(self.rectangle, "existing region with score: %i" % cur_score)
+            return
 
         if highscore>=100:
-            region = [r for r,s in scores.items() if s==highscore][0]
-            return setnewregion(region, "high score: %s", highscore)
+            region = next(iter(r for r,s in scores.items() if s==highscore))
+            setnewregion(region, "high score: %s", highscore)
+            return
 
-        #TODO:
+        #could do:
         # * re-add some scrolling detection: the region may have moved
         # * re-try with a higher "from_time" and a higher score threshold
 
@@ -536,7 +579,8 @@ class VideoSubregion(object):
             merged = merge_all(tuple(damage_count.keys()))
             score = score_region("merged", merged)
             if score>=110:
-                return setnewregion(merged, "merged all regions, score=%s", score)
+                setnewregion(merged, "merged all regions, score=%s", score)
+                return
 
         self.novideoregion("failed to identify a video region")
         self.last_scores = scores
